@@ -21,13 +21,16 @@
 
 #include <stdlib.h>
 
+#include <locale.h>
+
 #include <gtkmm/main.h>
+#include <giomm/volumemonitor.h>
+#include <giomm/init.h>
 #include <glib.h>
 #include <glib/gi18n.h>
 #include <gtk/gtk.h>
 #include <gdk/gdkx.h>
 #include <bacon-message-connection.h>
-#include <libgnomevfs/gnome-vfs.h>
 #include <gconf/gconf-client.h>
 #include <glibtop.h>
 #include <glibtop/close.h>
@@ -42,7 +45,7 @@
 #include "smooth_refresh.h"
 #include "util.h"
 #include "gconf-keys.h"
-
+#include "argv.h"
 
 
 ProcData::ProcData()
@@ -236,12 +239,6 @@ procman_data_new (GConfClient *client)
 
 	pd = ProcData::get_instance();
 
-	/* username is usually 8 chars long
-	   for caching, we create chunks of 128 chars */
-	pd->users = g_string_chunk_new(128);
-	/* push empty string */
-	g_string_chunk_insert_const(pd->users, "");
-
 	pd->config.width = gconf_client_get_int (client, "/apps/procman/width", NULL);
 	pd->config.height = gconf_client_get_int (client, "/apps/procman/height", NULL);
 	pd->config.show_tree = gconf_client_get_bool (client, "/apps/procman/show_tree", NULL);
@@ -370,7 +367,6 @@ procman_free_data (ProcData *procdata)
 {
 
 	proctable_free_table (procdata);
-	g_string_chunk_free(procdata->users);
 	delete procdata->smooth_refresh;
 }
 
@@ -595,14 +591,20 @@ cb_server (const gchar *msg, gpointer user_data)
 {
 	GdkWindow *window;
 	ProcData *procdata;
-	guint32 timestamp;
+	guint32 timestamp = 0;
 
 	window = gdk_get_default_root_window ();
 
 	procdata = *(ProcData**)user_data;
 	g_assert (procdata != NULL);
 
-	timestamp = strtoul(msg, NULL, 0);
+	procman_debug("cb_server(%s)", msg);
+	if (msg != NULL && procman::SHOW_SYSTEM_TAB_CMD == msg) {
+		procman_debug("Changing to PROCMAN_TAB_SYSINFO via bacon message");
+		gtk_notebook_set_current_page(GTK_NOTEBOOK(procdata->notebook), PROCMAN_TAB_SYSINFO);
+		cb_change_current_page(GTK_NOTEBOOK(procdata->notebook), PROCMAN_TAB_SYSINFO, procdata);
+	} else
+		timestamp = strtoul(msg, NULL, 0);
 
 	if (timestamp == 0)
 	{
@@ -616,20 +618,36 @@ cb_server (const gchar *msg, gpointer user_data)
 }
 
 
+
+
+static void
+mount_changed(const Glib::RefPtr<Gio::Mount>&)
+{
+  cb_update_disks(ProcData::get_instance());
+}
+
+
 static void
 init_volume_monitor(ProcData *procdata)
 {
-	GnomeVFSVolumeMonitor *mon;
-	mon = gnome_vfs_get_volume_monitor();
+  using namespace Gio;
+  using namespace Glib;
 
-	g_signal_connect(mon, "volume_mounted",
-			 G_CALLBACK(cb_volume_mounted_or_unmounted), procdata);
+  RefPtr<VolumeMonitor> monitor = VolumeMonitor::get();
 
-	g_signal_connect(mon, "volume_unmounted",
-			 G_CALLBACK(cb_volume_mounted_or_unmounted), procdata);
+  monitor->signal_mount_added().connect(sigc::ptr_fun(&mount_changed));
+  monitor->signal_mount_changed().connect(sigc::ptr_fun(&mount_changed));
+  monitor->signal_mount_removed().connect(sigc::ptr_fun(&mount_changed));
 }
 
 
+namespace procman
+{
+	const std::string SHOW_SYSTEM_TAB_CMD("SHOWSYSTAB");
+}
+
+
+
 int
 main (int argc, char *argv[])
 {
@@ -641,9 +659,23 @@ main (int argc, char *argv[])
 	bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
 	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
 	textdomain (GETTEXT_PACKAGE);
-
+	setlocale (LC_ALL, "");
+	
 	startup_timestamp = get_startup_timestamp();
 
+	Glib::OptionContext context;
+	context.set_summary(_("A simple process and system monitor."));
+	context.set_ignore_unknown_options(true);
+	procman::OptionGroup option_group;
+	context.set_main_group(option_group);
+
+	try {
+		context.parse(argc, argv);
+	} catch (const Glib::Error& ex) {
+		g_error("Arguments parse error : %s", ex.what().c_str());
+	}
+
+	Gio::init();
 	Gtk::Main kit(&argc, &argv);
 	procman_debug("post gtk_init");
 
@@ -660,6 +692,9 @@ main (int argc, char *argv[])
 
 		timestamp = g_strdup_printf ("%" G_GUINT32_FORMAT, startup_timestamp);
 
+		if (option_group.show_system_tab)
+			bacon_message_connection_send(conn, procman::SHOW_SYSTEM_TAB_CMD.c_str());
+
 		bacon_message_connection_send (conn, timestamp);
 
 		gdk_notify_startup_complete ();
@@ -678,7 +713,6 @@ main (int argc, char *argv[])
 	client = gconf_client_get_default ();
 	gconf_client_add_dir(client, "/apps/procman", GCONF_CLIENT_PRELOAD_NONE, NULL);
 
-	gnome_vfs_init ();
 	glibtop_init ();
 
 	procman_debug("end init");
@@ -696,6 +730,12 @@ main (int argc, char *argv[])
 
 	g_assert(procdata->app);
 			
+	if (option_group.show_system_tab) {
+		procman_debug("Starting with PROCMAN_TAB_SYSINFO by commandline request");
+		gtk_notebook_set_current_page(GTK_NOTEBOOK(procdata->notebook), PROCMAN_TAB_SYSINFO);
+		cb_change_current_page (GTK_NOTEBOOK(procdata->notebook), PROCMAN_TAB_SYSINFO, procdata);
+	}
+
  	gtk_widget_show(procdata->app);
        
 	procman_debug("begin gtk_main");
@@ -704,7 +744,6 @@ main (int argc, char *argv[])
 	procman_free_data (procdata);
 
 	glibtop_close ();
-	gnome_vfs_shutdown ();
 
 	return 0;
 }
