@@ -1,8 +1,14 @@
 #include <config.h>
 
 #include <glib.h>
+#include <gtk/gtk.h>
+#include <glibmm.h>
 #include <glib/gi18n.h>
 
+#include <libxml/parser.h>
+#include <libxml/xpath.h>
+#include <libxml/xpathInternals.h>
+
 #include <glibtop/fsusage.h>
 #include <glibtop/mountlist.h>
 #include <glibtop/mem.h>
@@ -15,8 +21,6 @@
 #include <math.h>
 #include <errno.h>
 
-#include "regex.h"
-
 #include <string>
 #include <vector>
 #include <fstream>
@@ -24,6 +28,7 @@
 #include <sys/utsname.h>
 
 #include "sysinfo.h"
+#include "procman.h"
 #include "util.h"
 
 
@@ -39,9 +44,9 @@ namespace {
   public:
     string hostname;
     string distro_name;
-    string distro_codename;
-    string distro_release; // numerical version
-
+    string distro_release;
+    string kernel;
+    string gnome_version;
     guint64 memory_bytes;
     guint64 free_space_bytes;
 
@@ -55,11 +60,30 @@ namespace {
       this->load_memory_info();
       this->load_disk_info();
       this->load_uname_info();
+      this->load_gnome_version();
     }
 
     virtual ~SysInfo()
     { }
 
+    virtual void set_distro_labels(GtkWidget* name, GtkWidget* release)
+    {
+      g_object_set(G_OBJECT(name),
+		   "label",
+		   ("<big><big><b>" + this->distro_name + "</b></big></big>").c_str(),
+		   NULL);
+
+
+      char* markup = g_strdup_printf(_("Release %s"), this->distro_release.c_str());
+
+      g_object_set(G_OBJECT(release),
+		   "label",
+		   markup,
+		   NULL);
+
+      g_free(markup);
+    }
+
   private:
 
     void load_memory_info()
@@ -125,8 +149,40 @@ namespace {
       uname(&name);
 
       this->hostname = name.nodename;
-      this->distro_name = name.sysname;
-      this->distro_release = name.release;
+      this->kernel = string(name.sysname) + ' ' + name.release;
+    }
+
+
+    void load_gnome_version()
+    {
+      xmlDocPtr document;
+      xmlXPathContextPtr context;
+      const string nodes[3] = { "string(/gnome-version/platform)",
+				"string(/gnome-version/minor)",
+				"string(/gnome-version/micro)" };
+      string values[3];
+
+      if (not (document = xmlParseFile(DATADIR "/gnome-about/gnome-version.xml")))
+	return;
+
+      if (not (context = xmlXPathNewContext(document)))
+	return;
+
+      for (size_t i = 0; i != 3; ++i)
+	{
+	  xmlXPathObjectPtr xpath;
+	  xpath = xmlXPathEvalExpression(BAD_CAST nodes[i].c_str(), context);
+	  
+	  if (xpath and xpath->type == XPATH_STRING)
+	    values[i] = reinterpret_cast<const char*>(xpath->stringval);
+
+	  xmlXPathFreeObject(xpath);
+	}
+
+      xmlXPathFreeContext(context);
+      xmlFreeDoc(document);
+
+      this->gnome_version = values[0] + '.' + values[1] + '.' + values[2];
     }
   };
 
@@ -159,17 +215,36 @@ namespace {
   {
   public:
     LSBSysInfo()
-      : re("^.+?:\\s*(.+)\\s*$")
+      : re(Glib::Regex::create("^.+?:\\s*(.+)\\s*$"))
     {
-      this->lsb_release();
+      // start();
     }
 
+    virtual void set_distro_labels(GtkWidget* name, GtkWidget* release)
+    {
+      this->name = name;
+      this->release = release;
+
+      this->start();
+    }
+
+
   private:
 
+    sigc::connection child_watch;
+    int lsb_fd;
+    GtkWidget* name;
+    GtkWidget* release;
+
     void strip_description(string &s) const
     {
-      // make a copy to avoid aliasing
-      this->re.PartialMatch(string(s), &s);
+      const GRegexMatchFlags flags = static_cast<GRegexMatchFlags>(0);
+      GMatchInfo* info = 0;
+
+      if (g_regex_match(this->re->gobj(), s.c_str(), flags, &info)) {
+	s = make_string(g_match_info_fetch(info, 1));
+	g_match_info_free(info);
+      }
     }
 
     std::istream& get_value(std::istream &is, string &s) const
@@ -179,7 +254,73 @@ namespace {
       return is;
     }
 
-    void lsb_release()
+
+    void read_lsb(Glib::Pid pid, int status)
+    {
+      this->child_watch.disconnect();
+
+      if (!WIFEXITED(status) or WEXITSTATUS(status) != 0) {
+	g_error("Child %d failed with status %d", int(pid), status);
+	return;
+      }
+
+      Glib::RefPtr<Glib::IOChannel> channel = Glib::IOChannel::create_from_fd(this->lsb_fd);
+      Glib::ustring content;
+
+      while (channel->read_to_end(content) == Glib::IO_STATUS_AGAIN)
+	;
+
+      channel->close();
+      Glib::spawn_close_pid(pid);
+
+      procman_debug("lsb_release output = '%s'", content.c_str());
+
+      string release, codename;
+      std::istringstream input(content);
+
+      this->get_value(input, this->distro_name)
+	and this->get_value(input, release)
+	and this->get_value(input, codename);
+
+      this->distro_release = release;
+      if (codename != "" && codename != "n/a")
+	this->distro_release += " (" + codename + ')';
+
+      this->SysInfo::set_distro_labels(this->name, this->release);
+    }
+
+
+    void start()
+    {
+      std::vector<string> argv(2);
+      argv[0] = "lsb_release";
+      argv[1] = "-irc";
+
+      Glib::SpawnFlags flags = Glib::SPAWN_DO_NOT_REAP_CHILD
+	| Glib::SPAWN_SEARCH_PATH
+	| Glib::SPAWN_STDERR_TO_DEV_NULL;
+
+      Glib::Pid child;
+
+      try {
+	Glib::spawn_async_with_pipes("/", // wd
+				     argv,
+				     flags,
+				     sigc::slot<void>(), // child setup
+				     &child,
+				     0, // stdin
+				     &this->lsb_fd); // stdout
+      } catch (Glib::SpawnError &e) {
+	g_error("g_spawn_async_with_pipes error: %s", e.what().c_str());
+	return;
+      }
+
+      sigc::slot<void,GPid, int> slot = sigc::mem_fun(this, &LSBSysInfo::read_lsb);
+      this->child_watch = Glib::signal_child_watch().connect(slot, child);
+    }
+
+
+    void sync_lsb_release()
     {
       char *out= 0;
       GError *error = 0;
@@ -190,12 +331,15 @@ namespace {
 				    0,
 				    &status,
 				    &error)) {
-
+	string release, codename;
 	if (!error and WIFEXITED(status) and WEXITSTATUS(status) == 0) {
 	  std::istringstream input(out);
 	  this->get_value(input, this->distro_name)
-	    and this->get_value(input, this->distro_release)
-	    and this->get_value(input, this->distro_codename);
+	    and this->get_value(input, release)
+	    and this->get_value(input, codename);
+	  this->distro_release = release;
+	  if (codename != "" && codename != "n/a")
+	    this->distro_release += " (" + codename + ')';
 	}
       }
 
@@ -206,7 +350,7 @@ namespace {
     }
 
   private:
-    const pcrecpp::RE re;
+    Glib::RefPtr<Glib::Regex> re;
   };
 
 
@@ -240,7 +384,7 @@ sysinfo_logo_expose (GtkWidget *widget,
 
   cr = gdk_cairo_create(widget->window);
 
-  cairo_translate(cr, event->area.x, event->area.y);
+  cairo_translate(cr, widget->allocation.x, widget->allocation.y);
 
   cairo_move_to(cr, X_PAD + RADIUS, Y_PAD);
   cairo_line_to(cr, X_PAD + LOGO_W - RADIUS, Y_PAD);
@@ -272,7 +416,60 @@ sysinfo_logo_expose (GtkWidget *widget,
   return FALSE;
 }
 
-GtkWidget *
+static GtkWidget*
+add_section(GtkBox *vbox , const char * title, int num_row, int num_col, GtkWidget **out_frame)
+{
+  GtkWidget *table;
+
+  GtkWidget *frame = gtk_frame_new(title);
+  gtk_frame_set_label_align(GTK_FRAME(frame), 0.0, 0.5);
+  gtk_label_set_use_markup(
+			   GTK_LABEL(gtk_frame_get_label_widget(GTK_FRAME(frame))),
+			   TRUE
+			   );
+  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_NONE);
+  gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
+
+  GtkWidget *alignment = gtk_alignment_new(0.5, 0.5, 1.0, 1.0);
+  gtk_alignment_set_padding(GTK_ALIGNMENT(alignment), 0, 0, 12, 0);
+  gtk_container_add(GTK_CONTAINER(frame), alignment);
+
+  table = gtk_table_new(num_row, num_col, FALSE);
+  gtk_table_set_row_spacings(GTK_TABLE(table), 6);
+  gtk_table_set_col_spacings(GTK_TABLE(table), 6);
+  gtk_container_set_border_width(GTK_CONTAINER(table), 6);
+  gtk_container_add(GTK_CONTAINER(alignment), table);
+
+  if(out_frame)
+    *out_frame = frame;
+
+  return table;
+}
+
+
+static GtkWidget* 
+add_row(GtkTable * table, const char * label, const char * value, int row)
+{
+  GtkWidget *header = gtk_label_new(label);
+  gtk_misc_set_alignment(GTK_MISC(header), 0.0, 0.5);
+  gtk_table_attach(
+		   table, header,
+		   0, 1, row, row + 1,
+		   GTK_FILL, GTK_FILL, 0, 0
+		   );
+
+  GtkWidget *label_widget = gtk_label_new(value);
+  gtk_misc_set_alignment(GTK_MISC(label_widget), 0.0, 0.5);
+  gtk_table_attach(
+		   table, label_widget,
+		   1, 2, row, row + 1,
+		   GTK_FILL, GTK_FILL, 0, 0
+		   );
+  return label_widget;
+}
+
+
+static GtkWidget *
 procman_create_sysinfo_view(void)
 {
   GtkWidget *hbox;
@@ -284,18 +481,16 @@ procman_create_sysinfo_view(void)
 
   GtkWidget *distro_frame;
   GtkWidget *distro_release_label;
+  GtkWidget *distro_table;
 
-  GtkWidget *hardware_frame;
   GtkWidget *hardware_table;
   GtkWidget *memory_label;
   GtkWidget *processor_label;
 
-  GtkWidget *disk_space_frame;
   GtkWidget *disk_space_table;
   GtkWidget *disk_space_label;
 
   GtkWidget *header;
-  GtkWidget *alignment;
 
   gchar *markup;
 
@@ -334,144 +529,108 @@ procman_create_sysinfo_view(void)
 
   /* distro section */
 
-  markup = g_strdup_printf("<big><big><b>%s</b></big></big>",
-    data->distro_name.c_str());
-  distro_frame = gtk_frame_new(markup);
-  g_free(markup);
-  gtk_frame_set_label_align(GTK_FRAME(distro_frame), 0.0, 0.5);
-  gtk_label_set_use_markup(
-			   GTK_LABEL(gtk_frame_get_label_widget(GTK_FRAME(distro_frame))),
-			   TRUE
-			   );
-  gtk_frame_set_shadow_type(GTK_FRAME(distro_frame), GTK_SHADOW_NONE);
-  gtk_box_pack_start(GTK_BOX(vbox), distro_frame, FALSE, FALSE, 0);
+  unsigned table_size = 2;
+  if (data->gnome_version != "")
+     table_size++;
+  distro_table = add_section(GTK_BOX(vbox), "???", table_size, 1, &distro_frame);
 
-  alignment = gtk_alignment_new(0.5, 0.5, 1.0, 1.0);
-  gtk_alignment_set_padding(GTK_ALIGNMENT(alignment), 0, 0, 12, 0);
-  gtk_container_add(GTK_CONTAINER(distro_frame), alignment);
+  unsigned table_count = 0;
 
-  if (data->distro_codename != "")
-    markup = g_strdup_printf(
-			     _("Release %s (%s)"),
-			     data->distro_release.c_str(),
-			     data->distro_codename.c_str());
-  else
-    markup = g_strdup_printf(
-			     _("Release %s"),
-			     data->distro_release.c_str());
-  distro_release_label = gtk_label_new(markup);
-  g_free(markup);
+  distro_release_label = gtk_label_new("???");
   gtk_misc_set_alignment(GTK_MISC(distro_release_label), 0.0, 0.5);
-  gtk_misc_set_padding(GTK_MISC(distro_release_label), 6, 6);
-  gtk_container_add(GTK_CONTAINER(alignment), distro_release_label);
+  gtk_table_attach(
+		   GTK_TABLE(distro_table), distro_release_label,
+		   0, 1, table_count, table_count+1,
+		   GTK_FILL, GTK_FILL, 0, 0
+		   );
+  table_count++;
+  data->set_distro_labels(gtk_frame_get_label_widget(GTK_FRAME(distro_frame)), distro_release_label);
+
+  markup = g_strdup_printf(_("Kernel %s"), data->kernel.c_str());
+  header = gtk_label_new(markup);
+  g_free(markup);
+  gtk_misc_set_alignment(GTK_MISC(header), 0.0, 0.5);
+  gtk_table_attach(
+		   GTK_TABLE(distro_table), header,
+		   0, 1, table_count, table_count + 1,
+		   GTK_FILL, GTK_FILL, 0, 0
+		   );
+  table_count++;
+
+  if (data->gnome_version != "")
+    {
+      markup = g_strdup_printf(_("GNOME %s"), data->gnome_version.c_str());
+      header = gtk_label_new(markup);
+      g_free(markup);
+      gtk_misc_set_alignment(GTK_MISC(header), 0.0, 0.5);
+      gtk_table_attach(
+		       GTK_TABLE(distro_table), header,
+		       0, 1, table_count, table_count + 1,
+		       GTK_FILL, GTK_FILL, 0, 0
+		       );
+      table_count++;
+    }
 
   /* hardware section */
 
   markup = g_strdup_printf(_("<b>Hardware</b>"));
-  hardware_frame = gtk_frame_new(markup);
+  hardware_table = add_section(GTK_BOX(vbox), markup, data->processors.size(), 2, NULL);
   g_free(markup);
-  gtk_frame_set_label_align(GTK_FRAME(hardware_frame), 0.0, 0.5);
-  gtk_label_set_use_markup(
-			   GTK_LABEL(gtk_frame_get_label_widget(GTK_FRAME(hardware_frame))),
-			   TRUE
-			   );
-  gtk_frame_set_shadow_type(GTK_FRAME(hardware_frame), GTK_SHADOW_NONE);
-  gtk_box_pack_start(GTK_BOX(vbox), hardware_frame, FALSE, FALSE, 0);
-
-  alignment = gtk_alignment_new(0.5, 0.5, 1.0, 1.0);
-  gtk_alignment_set_padding(GTK_ALIGNMENT(alignment), 0, 0, 12, 0);
-  gtk_container_add(GTK_CONTAINER(hardware_frame), alignment);
-
-  hardware_table = gtk_table_new(data->processors.size(), 2, FALSE);
-  gtk_table_set_row_spacings(GTK_TABLE(hardware_table), 6);
-  gtk_table_set_col_spacings(GTK_TABLE(hardware_table), 6);
-  gtk_container_set_border_width(GTK_CONTAINER(hardware_table), 6);
-  gtk_container_add(GTK_CONTAINER(alignment), hardware_table);
 
+  markup = procman::format_size(data->memory_bytes);
+  memory_label = add_row(GTK_TABLE(hardware_table), _("Memory:"),
+                         markup, 0);
+  g_free(markup);
   header = gtk_label_new(_("Memory:"));
-  gtk_misc_set_alignment(GTK_MISC(header), 0.0, 0.5);
-  gtk_table_attach(
-		   GTK_TABLE(hardware_table), header,
-		   0, 1, 0, 1,
-		   GTK_FILL, GTK_FILL, 0, 0
-		   );
-
-  markup = SI_gnome_vfs_format_file_size_for_display(data->memory_bytes);
-  memory_label = gtk_label_new(markup);
-  g_free(markup);
-  gtk_misc_set_alignment(GTK_MISC(memory_label), 0.0, 0.5);
-  gtk_table_attach(
-		   GTK_TABLE(hardware_table), memory_label,
-		   1, 2, 0, 1,
-		   GTK_FILL, GTK_FILL, 0, 0
-		   );
 
   for (guint i = 0; i < data->processors.size(); ++i) {
+    const gchar * t;
     if (data->processors.size() > 1) {
       markup = g_strdup_printf(_("Processor %d:"), i);
-      header = gtk_label_new(markup);
-      g_free(markup);
+      t = markup;
+    }
+    else {
+      markup = NULL;
+      t = _("Processor:");
     }
-    else
-      header = gtk_label_new(_("Processor:"));
 
-    gtk_misc_set_alignment(GTK_MISC(header), 0.0, 0.5);
-    gtk_table_attach(
-		     GTK_TABLE(hardware_table), header,
-		     0, 1, 1 + i, 2 + i,
-		     GTK_FILL, GTK_FILL, 0, 0
-		     );
+    processor_label = add_row(GTK_TABLE(hardware_table), t,
+                              data->processors[i].c_str(), 1 + i);
 
-    processor_label = gtk_label_new(data->processors[i].c_str());
-    gtk_misc_set_alignment(GTK_MISC(processor_label), 0.0, 0.5);
-    gtk_table_attach(
-		     GTK_TABLE(hardware_table), processor_label,
-		     1, 2, 1 + i, 2 + i,
-		     GTK_FILL, GTK_FILL, 0, 0
-		     );
+    if(markup)
+      g_free(markup);
   }
 
   /* disk space section */
 
   markup = g_strdup_printf(_("<b>System Status</b>"));
-  disk_space_frame = gtk_frame_new(markup);
+  disk_space_table = add_section(GTK_BOX(vbox), markup, 1, 2, NULL);
   g_free(markup);
-  gtk_frame_set_label_align(GTK_FRAME(disk_space_frame), 0.0, 0.5);
-  gtk_label_set_use_markup(
-			   GTK_LABEL(gtk_frame_get_label_widget(GTK_FRAME(disk_space_frame))),
-			   TRUE
-			   );
-  gtk_frame_set_shadow_type(GTK_FRAME(disk_space_frame), GTK_SHADOW_NONE);
-  gtk_box_pack_start(GTK_BOX(vbox), disk_space_frame, FALSE, FALSE, 0);
 
-  alignment = gtk_alignment_new(0.5, 0.5, 1.0, 1.0);
-  gtk_alignment_set_padding(GTK_ALIGNMENT(alignment), 0, 0, 12, 0);
-  gtk_container_add(GTK_CONTAINER(disk_space_frame), alignment);
-
-  disk_space_table = gtk_table_new(1, 2, FALSE);
-  gtk_table_set_row_spacings(GTK_TABLE(disk_space_table), 6);
-  gtk_table_set_col_spacings(GTK_TABLE(disk_space_table), 6);
-  gtk_container_set_border_width(GTK_CONTAINER(disk_space_table), 6);
-  gtk_container_add(GTK_CONTAINER(alignment), disk_space_table);
-
-  header = gtk_label_new(_("Available disk space:"));
-  gtk_misc_set_alignment(GTK_MISC(header), 0.0, 0.5);
-  gtk_table_attach(
-		   GTK_TABLE(disk_space_table), header,
-		   0, 1, 0, 1,
-		   GTK_FILL, GTK_FILL, 0, 0
-		   );
-
-  markup = SI_gnome_vfs_format_file_size_for_display(data->free_space_bytes);
-  disk_space_label = gtk_label_new(markup);
+  markup = procman::format_size(data->free_space_bytes);
+  disk_space_label = add_row(GTK_TABLE(disk_space_table), 
+                             _("Available disk space:"), markup,
+                             0);
   g_free(markup);
-  gtk_misc_set_alignment(GTK_MISC(disk_space_label), 0.0, 0.5);
-  gtk_table_attach(
-		   GTK_TABLE(disk_space_table), disk_space_label,
-		   1, 2, 0, 1,
-		   GTK_FILL, GTK_FILL, 0, 0
-		   );
 
   return hbox;
 }
 
+
+
+namespace procman
+{
+  void build_sysinfo_ui()
+  {
+    static GtkWidget* ui;
+
+    if (!ui) {
+      ProcData* procdata = ProcData::get_instance();
+      ui = procman_create_sysinfo_view();
+      GtkBox* box = GTK_BOX(gtk_notebook_get_nth_page(GTK_NOTEBOOK(procdata->notebook),
+						      PROCMAN_TAB_SYSINFO));
+      gtk_box_pack_start(box, ui, TRUE, TRUE, 0);
+      gtk_widget_show_all(ui);
+    }
+  }
+}
